#!/usr/bin/env python

import os
import copy
import time
from urllib.parse import urlparse
import requests
import logging

from nextcodecli.utils import find_aws_credentials, get_logger

logging.getLogger('botocore').setLevel(logging.WARNING)


def find_job(session, job_id):
    log = get_logger()
    job_info = None
    job_url = None
    if str(job_id) == 'latest':
        if not session.current_user:
            log.error("No user in session")
            return None

        user_name = session.current_user['preferred_username']
        resp = session.get(
            session.url_from_endpoint('jobs'),
            json={
                'user_name': user_name,
                'field': ['job_id', 'links'],
                'status': 'ALL',
                'limit': 1,
            },
        )
        resp.raise_for_status()
        jobs = resp.json()['jobs']
        if len(jobs) > 0:
            job_info = jobs[0]
            job_url = job_info['links']['self']
            log.info("Found latest job for %s: %s", user_name, job_url)
        else:
            log.warning("Found no jobs for %s", user_name)
            return None

    else:
        job_id = int(job_id)
        job_endpoint = session.url_from_endpoint('jobs')
        log.debug('job_endpoint=%s, job_id=%s' % (job_endpoint, job_id))
        resp = session.get(job_endpoint, json={'job_id': job_id, 'field': ['job_id', 'links']})
        resp.raise_for_status()
        jobs = resp.json()['jobs']
        if not jobs:
            log.warning("Job '%s' not found" % job_id)
            return None
        job_info = jobs[0]
        job_url = job_info['links']['self']

    fields = []
    # Now get the actual resource
    try:
        resp = session.get(job_url, data={'field': [fields]})
        resp.raise_for_status()
        return resp.json()
    except Exception as e:
        log.error("Error getting job info from %s: %s", job_url, e)
        log.error("Just returning basic job info!")
        return job_info


def download_output_file(folder_name, url):
    log = get_logger()
    url_parts = urlparse(url)
    file_name = url_parts.path.split('/')[-1]
    log.info("Downloading '%s' to %s..." % (file_name, folder_name))
    path = os.path.expanduser(folder_name)
    try:
        os.makedirs(path)
    except Exception:
        pass
    local_filename = os.path.join(path, file_name)
    r = requests.get(url, stream=True)
    with open(local_filename, 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)
                # f.flush() commented by recommendation from J.F.Sebastian
    return local_filename


def post_job(
    session,
    manifest,
    build_source=None,
    build_context=None,
    pipeline_definition=None,
    force_start=False,
    is_test=False,
):
    log = get_logger()

    # ! TODO: Check json
    pipeline_name = manifest['pipeline_name']

    url = session.url_from_endpoint('jobs')
    log.info("Running pipeline '%s' against endpoint '%s'...", pipeline_name, url)
    log.info("Samples: %s" % (", ".join([s['name'] for s in manifest['samples']])))

    # copy most things from the manifest to the post body
    skip_keys = ['samples']
    job_description = copy.deepcopy(manifest)
    for k in skip_keys:
        try:
            del job_description[k]
        except Exception:
            pass

    # if we do not have iam_keys in the job description, assume we have iam_profiles and pull
    # the keys from the user's aws credentials file
    if 'iam_keys' not in job_description:
        iam_keys = []
        for profile_name, profile in manifest['iam_profiles'].items():
            credentials = find_aws_credentials(profile)
            iam_key = {'profile': profile_name}
            iam_key.update(credentials)
            iam_keys.append(iam_key)
        job_description['iam_keys'] = iam_keys
    try:
        del job_description['iam_profiles']
    except KeyError:
        pass

    # allow user to upload etl-pipelines package (!! dev mode)
    if build_source:
        log.info("Using build source %s with context '%s'", build_source, build_context)
        job_description['build_source'] = build_source
        job_description['build_context'] = build_context

    # allow user to override the pipeline definition with local file (!! dev mode)
    if pipeline_definition is not None:
        job_description['pipeline_definition'] = pipeline_definition
    responses = []

    # now go through the samples and post a separate job for each one
    for sample in manifest['samples']:
        sample_name = sample['name']
        if 'input' in sample:
            input_files = sample.get('input')
        # backwards compatibility
        elif 'files' in sample:
            input_files = {'samples': sample['files']}
            log.warning(
                "Backwards compatibility for 'files' in manifest. Please fix your manifest file"
            )
        input_list = []
        for key_name, files in input_files.items():
            input_list.append({'key': key_name, 'files': files})
        this_job = copy.deepcopy(job_description)

        # allow override of all parameters for each sample
        for key in this_job.keys():
            if key in sample:
                val = sample[key]
                if isinstance(val, dict):
                    this_job[key].update(val)
                else:
                    this_job[key] = val

        this_job['sample_name'] = sample_name
        this_job['input'] = input_list
        if is_test:
            this_job['test'] = True

        log.info("Creating job for sample '%s'...", sample_name)
        resp = session.post(url, json=this_job)
        if resp.status_code == requests.codes.created:
            log.info("Success: Job %s has been created", resp.json()['job_id'])
        else:
            log.error("Failure: %s" % resp.text)
        if not is_test and force_start and resp.status_code == requests.codes.created:
            links = session.links(resp.json())
            job_url = links['self']
            instances_url = links['instances']
            log.info("Force job start...")
            instance_resp = session.post(instances_url, json={})
            if not instance_resp.status_code == requests.codes.created:
                log.error("Failed to launch instance: %s" % instance_resp.text)
            resp = session.get(job_url)
        responses.append(resp)
        # do not flood the server
        time.sleep(0.1)
    return responses



#!/usr/bin/env python

import datetime
import collections

import click
from tabulate import tabulate
import dateutil.parser

status_to_color = {
    'COMPLETED': 'green',
    'QUEUED': 'yellow',
    'INITIALIZING': 'yellow',
    'LAUNCHING': 'yellow',
    'RUNNING': 'yellow',
    'FAILED': 'red',
    'FINISHED': 'green',
}


def print_filelist(files, details=False):
    if not files:
        click.echo("Not found")
        return
    table = []
    for file in files:
        if details:
            sz = int(file.get('file_size', -1))
            if sz > 1024 * 1024:
                sz = "%.0f MB" % (int(sz) // 1024 // 1024)
            elif sz > 1024:
                sz = "%.0f KB" % (int(sz) // 1024)
            elif sz > 0:
                sz = "%.0f B" % sz
            else:
                sz = click.style('NOT FOUND', fg='yellow')

            table.append((file['file_name'], file['remote_path'], sz))
        else:
            table.append((file['file_name'], file['remote_path'], ""))
    tbl = tabulate(table, headers=['file_name', 'remote path', ''])
    click.echo(tbl)


def get_steps_table(job, color=True):
    step_list = []
    pending_steps = collections.defaultdict(int)
    for step in job.get('steps', []):
        dt_start = ''
        duration = '-'
        if step.get('date_started'):
            dt_start = dateutil.parser.parse(step['date_started'])
            dt_start = dt_start.replace(tzinfo=None)

            if step['date_completed']:
                dt_end = dateutil.parser.parse(step['date_completed'])
            else:
                dt_end = datetime.datetime.utcnow()
            dt_end = dt_end.replace(tzinfo=None)
            duration = (dt_end - dt_start).total_seconds()
            hours, remainder = divmod(duration, 3600)
            minutes, seconds = divmod(remainder, 60)
            duration = '%02d:%02d:%02d' % (int(hours), int(minutes), int(seconds))
            dt_start = dt_start.strftime("%H:%M")
            status = step['status']
            if color:
                col = status_to_color.get(status, '')
                status = click.style(status, fg=col, bold=True)
            step_list.append(
                (step['step_name'], step.get('step_identifier', ''), dt_start, duration, status)
            )
        else:
            pending_steps[step['step_name']] += 1

    for step_name in sorted(pending_steps):
        count = pending_steps[step_name]
        txt = step_name
        if count > 1:
            txt += " (x%s)" % count
        step_list.append((txt, '', '', '', 'PENDING'))
    steps_txt = tabulate(step_list, headers=['step', 'identifier', 'started', 'duration', 'status'])
    return steps_txt



#!/usr/bin/env python

import sys
import click
from click import echo, secho, pass_context, command, argument, option
import dateutil
from tabulate import tabulate
import datetime
import time
import subprocess
from urllib.parse import urlsplit, urljoin

from nextcode.services.workflow import INIT_STATUSES
from nextcode.exceptions import ServerError
from nextcode.config import get_default_profile

from nextcodecli.utils import dumps, print_tab, print_warn, print_error, output_pretty_json, abort
from nextcodecli.commands.workflow import status_to_color
from nextcodecli.workflow.cmdutils import get_processes_table, process_status_to_color


def check_for_broken_job(job):
    status = job.status
    if status != 'PENDING':
        return False
    diff = (datetime.datetime.utcnow() - job.submit_date).total_seconds()
    if diff >= 60.0:
        secho(
            "\nJob has been in status '%s' for %.0f seconds. Something might be wrong."
            % (status, diff),
            fg='yellow',
        )
        cmd = ['nextcode', 'workflow', 'job', str(job.job_id), 'inspect']
        echo("\nExecuting %s..." % ' '.join(cmd))
        subprocess.call(cmd)
        return True
    return False


@command(
    cls=click.Group,
    invoke_without_command=True,
    help="View or manage individual jobs. Call with [job_id] or latest to "
    "view your latest submitted job",
)
@argument('job_id', required=True)
@pass_context
def job(ctx, job_id):
    if ctx.obj.job is None:
        ctx.obj.job = ctx.obj.service.find_job(job_id)
        ctx.obj.job_id = ctx.obj.job.job_id
    if ctx.invoked_subcommand is None:
        ctx.invoke(view)
        echo(job.get_help(ctx))


@job.command(help="Investigate a running job for problems")
@option('--raw', 'is_raw', is_flag=True, help='Dump raw json response from endpoint')
@pass_context
def inspect(ctx, is_raw):
    workflow_job = ctx.obj.job
    try:
        curr_status = workflow_job.inspect()
    except ServerError as e:
        abort(e)

    if is_raw:
        echo(dumps(curr_status))
        return
    print_tab('Job id', workflow_job.job_id)
    job_status = workflow_job.status or 'N/A'
    print_tab('Job status', job_status)
    print_tab('Status message', workflow_job.status_message)
    latest_event = curr_status['latest_event']
    event_too_old = False
    if latest_event:
        dt = dateutil.parser.parse(latest_event)
        dt = dt.replace(tzinfo=None)
        diff = datetime.datetime.utcnow() - dt
        diff_txt = " (%s ago)" % str(diff).split('.')[0]
        if diff < datetime.timedelta(minutes=5):
            col = 'green'
            event_too_old = False
        else:
            col = 'red'
            event_too_old = True
        diff_txt = click.style(diff_txt, fg=col)
        latest_event += diff_txt
    print_tab('Latest event', latest_event)
    print_tab('Executor Pod name', workflow_job.details.get('pod_name'))
    executor_pod = None
    for pod in curr_status['pods']:
        if pod['metadata']['labels']['app-name'] == 'nextflow-executor':
            executor_pod = pod
            break
    if not executor_pod:
        if job_status in ('PENDING', 'STARTED'):
            secho("Nextflow executor pod not Found. This job is not running!", fg='red')
    else:
        status = executor_pod['status']
        container_statuses = status['container_statuses']
        pod_state = {}
        if container_statuses:
            pod_state = container_statuses[0]['state']
        node_name = executor_pod['spec']['node_name']
        print_tab('Executor node name', node_name)
        secho("\nExecutor Pod State:", bold=True)
        _keys = ['running', 'terminated', 'waiting']
        keys = []
        for k in _keys:
            if pod_state.get(k):
                keys.append(k)
        output_pretty_json(pod_state, keys=keys)
        echo(
            "Hint: You can view latest logs from this pod with: 'nextcode workflow job {} logs pod'".format(
                workflow_job.job_id
            )
        )

    secho("\nProcess Pod State:", bold=True)
    table = []
    for pod in curr_status['pods']:
        if pod == executor_pod:
            continue
        container_status = None
        pod_status = pod['status']
        container_statuses = pod_status['container_statuses']
        if container_statuses:
            pod_state = container_statuses[0]['state']
        else:
            pod_state = {}

        for k, v in pod_state.items():
            if v:
                container_status = k
                if k.lower() == 'terminated':
                    container_status = "%s (%s at %s)" % (k, v['reason'], v['finished_at'])
        p = [
            pod['metadata']['name'],
            container_status,
            pod['spec']['node_name'],
            pod['metadata']['labels'].get('taskName', 'N/A'),
        ]
        table.append(p)
    if not table:
        echo("No process pods found")
    else:
        tbl = tabulate(table, headers=['name', 'state', 'node', 'process'])
        echo(tbl)
        echo(
            "\nHint: You can find the corresponding processes with: 'nextcode workflow job {} processes -o wide'".format(
                workflow_job.job_id
            )
        )

    if job_status == 'COMPLETED':
        secho("\nThis job appears to have completed successfully", fg='green', bold=True)
    elif job_status in ('STARTED', 'PENDING'):
        if not executor_pod:
            secho("\nThis job appears to be broken", fg='red', bold=True)
        else:
            try:
                if executor_pod['status']['container_statuses'][0]['state']['running']:
                    if event_too_old:
                        secho(
                            "\nJob appears to be running but it has been a long time since last heartbeat was received",
                            fg='red',
                            bold=True,
                        )
                    else:
                        secho("\nThis job appears to be running properly", fg='green', bold=True)
                else:
                    secho("\nThis job appears to be broken", fg='red', bold=True)
            except Exception:
                secho("\nThis job appears to be broken", fg='red', bold=True)
    else:
        secho("\nJob is not in running state (%s)" % job_status, bold=True)

    if executor_pod and job_status == 'PENDING':
        cmd = ['nextcode', 'workflow', 'job', str(workflow_job.job_id), 'logs', 'pod']
        echo("\nExecuting %s..." % ' '.join(cmd))
        subprocess.call(cmd)
        return


@job.command(help="View information about the job")
@option('--raw', 'is_raw', is_flag=True, help='Dump raw json response from endpoint')
@pass_context
def view(ctx, is_raw):
    workflow_job = ctx.obj.job
    if is_raw:
        echo(dumps(workflow_job.job))
        return

    print_tab('job_id', workflow_job.job_id)
    pipeline_name = workflow_job.pipeline_name
    if not pipeline_name:
        pipeline_name = "(custom)"
    print_tab('pipeline_name', pipeline_name)
    print_tab('description', workflow_job.description)
    submit_date = workflow_job.submit_date
    print_tab('submitted', submit_date.strftime('%Y-%m-%d %H:%M:%S'))
    complete_date = workflow_job.complete_date or datetime.datetime.utcnow()
    if workflow_job.complete_date:
        print_tab('completed', complete_date.strftime('%Y-%m-%d %H:%M:%S'))
    print_tab('duration', str(complete_date - submit_date).split('.')[0])

    print_tab('nextflow ID', workflow_job.run_id)
    print_tab(
        'nextflow Pod',
        "%s (%s)"
        % (workflow_job.details.get('pod_name'), workflow_job.details.get('pod_namespace')),
    )
    print_tab('script_name', workflow_job.script_name)
    print_tab('revision', (workflow_job.revision or '(latest)'))
    print_tab('user', workflow_job.user_name)
    status = workflow_job.status or 'N/A'
    col = status_to_color.get(status, '')
    st = str(status)
    if workflow_job.status_message:
        st += " - {}".format(workflow_job.status_message)
    st = click.style(st, fg=col, bold=True)
    print_tab('status', st)

    txt = ", ".join(
        [
            "{}: {}".format(k.capitalize(), v)
            for k, v in workflow_job.details.get('process_stats', {}).items()
        ]
    )
    txt = "{}".format(txt)
    print_tab('processes', txt)

    if status in ('STARTED', 'ERROR'):
        limit = 50
        processes = workflow_job.processes(limit=limit)
        echo("Running Processes:")
        echo(get_processes_table(processes))
        if len(processes) == limit:
            secho(
                "Note: Only showing the latest {} non-completed processes".format(limit), bold=True
            )
        for p in processes:
            if p['status'] == 'FAILED':
                cmd = "nextcode workflow job %s process %s view|pod|logs" % (
                    workflow_job.job_id,
                    p['process_id'],
                )
                echo("Inspect failed process with: %s" % click.style(cmd, bold=True))

    check_for_broken_job(workflow_job)


@job.command(help="Get nextflow events for this job")
@option('--raw', 'is_raw', is_flag=True, help='Dump raw json response from endpoint')
@option('-e', '--event', 'event_name', help='Filter events on name')
@option('-p', '--process', 'process_name', help='Filter events on process')
@option('-n', '--num', default=50, help='Maximum number of events to return')
@pass_context
def events(ctx, is_raw, event_name, process_name, num):
    events = ctx.obj.job.events(num)
    if event_name:
        events = [e for e in events if e['event'].startswith(event_name)]
    if process_name:
        events = [
            e
            for e in events
            if e and (e.get('trace', {}) or {}).get('name', '').startswith(process_name)
        ]

    if is_raw:
        print(dumps(events))
        return

    headers = ['first_event_date', 'last_event_date', 'event', 'process', 'status']
    table = []
    for event in events:
        dt = dateutil.parser.parse(event['event_date'])
        try:
            last_dt = dateutil.parser.parse(event['last_event_date'])
        except Exception:
            last_dt = '-'
        row = [
            dt,
            last_dt,
            event['event'],
            (event.get('trace') or {}).get('name'),
            ((event.get('trace') or {}).get('status') or ""),
        ]
        if (event['num_events'] or 1) > 1:
            row[-1] += ' ( x%s )' % event['num_events']
        table.append(row)

    tbl = tabulate(table, headers=headers)
    echo(tbl)


@job.command(help="Retry a failed job.")
@pass_context
def resume(ctx):
    workflow_job = ctx.obj.job
    job_id = workflow_job.job_id
    echo("Resuming failed job %s..." % job_id)
    try:
        workflow_job.resume()
    except ServerError as e:
        abort(e)
    cmd = ['nextcode', 'workflow', 'job', str(job_id), 'watch']
    echo("Executing %s..." % ' '.join(cmd))
    subprocess.call(cmd)


@job.command(help="View logs associated with a job. Choose the log group to stream.")
@argument('log_group', default='')
@option('--filter', 'log_filter', help='Filter logs on text')
@pass_context
def logs(ctx, log_group, log_filter):
    workflow_job = ctx.obj.job
    if not log_group:
        log_groups = workflow_job.log_groups()
        echo("The following log groups are available:\n  %s" % ("\n  ".join(log_groups.keys())))
        sys.exit(1)
    try:
        logs = workflow_job.logs(log_group, log_filter)
    except Exception as ex:
        abort(ex)
    if not logs:
        print_warn("No logs found")
        return

    echo(logs)
    echo("\n%s loglines found" % (logs.count('\n')))


@job.command(help="Information about Nextflow processes for this job")
@option('--raw', 'is_raw', is_flag=True, help='Dump raw json response from endpoint')
@option('-p', '--process', 'process_name', help='Only show a particular process')
@option('-n', '--num', 'limit', default=100, help='Number of processes to show')
@option('-a', '--all', 'is_all', is_flag=True, default=False, help='Also show completed processes')
@option('-s', '--status', 'status', default=None, help='Filter process status')
@option(
    '-o',
    '--output',
    type=click.Choice(['normal', 'wide', 'json']),
    default='normal',
    help='Format output',
)
@pass_context
def processes(ctx, is_raw, process_name, output, limit, is_all, status):
    workflow_job = ctx.obj.job
    processes = workflow_job.processes(limit=limit, is_all=is_all, status=status)
    if process_name:
        processes = [p for p in processes if p['name'].startswith(process_name)]
    is_raw = is_raw or (output == 'json')
    if is_raw:
        echo(dumps(processes))
        return

    txt = click.style(
        ", ".join(
            [
                "{}: {}".format(k.capitalize(), v)
                for k, v in workflow_job.details.get('process_stats', {}).items()
            ]
        ),
        bold=True,
    )
    txt = "Processes by status: {}".format(txt)
    echo(txt)

    if len(processes) == 0 and not is_all:
        secho(
            "No running processes found. Use the '--all' flag to show completed processes",
            bold=True,
        )
        return
    if not is_all:
        echo("\nNon-completed processes:")
    else:
        echo("\nProcesses:")
    echo(get_processes_table(processes, is_wide=(output == 'wide')))
    if len(processes) == limit:
        msg = "Note: Only showing the latest {} non-completed processes".format(limit)
        if is_all or status:
            msg = "Note: Only showing the latest {} processes".format(limit)
        secho(msg, bold=True)


@job.command(
    cls=click.Group,
    invoke_without_command=True,
    help="Information about a specific nextflow process",
)
@argument('process_id', required=True)
@pass_context
def process(ctx, process_id):
    processes = ctx.obj.job.processes(process_id)
    if not processes:
        abort("Process %s not found in job %s" % (process_id, ctx.obj.job.job_id))
    ctx.obj.process = processes[0]

    if ctx.invoked_subcommand is None:
        ctx.invoke(view)
        echo(job.get_help(ctx))


@process.command('view', help="View nextflow process details")
@option('--raw', 'is_raw', is_flag=True, help='Dump raw json response from endpoint')
@pass_context
def process_view(ctx, is_raw):
    process = ctx.obj.process
    if is_raw:
        echo(dumps(process))
        return

    try:
        node_name = process['details']['node']['metadata']['name']
    except Exception:
        node_name = 'N/A'
    print_tab('job_id', process['job_id'])
    print_tab('process_id', process['process_id'])
    print_tab('name', process['name'])
    print_tab('pod', process['native_id'])
    print_tab('node', node_name)
    print_tab('process_hash', process['process_hash'])
    status = click.style(
        process['status'], fg=process_status_to_color.get(process['status'], 'white')
    )
    print_tab('status', status)
    print_tab('submit_date', process['submit_date'])
    print_tab('complete_date', process['complete_date'])
    print_tab('status_date', process['status_date'])
    print_tab('exit_code', process['exit_code'])
    print_tab('container', process['container'])


@process.command('logs', help="View pod logs for nextflow process (if the pod is still available)")
@pass_context
def process_logs(ctx):
    process = ctx.obj.process
    try:
        logs = ctx.obj.session.get(process['links']['logs'])
    except ServerError as e:
        abort(e)

    print("Logs from pod '%s':" % process['native_id'])
    print(logs.text)


@process.command('pod', help="View saved pod information for nextflow process from the process run")
@option('--raw', 'is_raw', is_flag=True, help='Dump raw json response from endpoint')
@pass_context
def process_pod(ctx, is_raw):
    process = ctx.obj.process
    if not process['details']:
        abort(
            "Process has no associated information. This could mean that the process has not started"
        )
    pod = (process['details'] or {}).get('pod', {})
    if is_raw:
        echo(dumps(pod))
        return

    echo("Showing pod information from when the process was run...  ")
    keys = [
        'name',
        'namespace',
        'creation_timestamp',
        'labels',
        'image',
        'state',
        'resources.limits',
    ]
    output_pretty_json(pod, keys=keys)


@process.command(
    'node', help="View saved node information for nextflow process from the process run"
)
@option('--raw', 'is_raw', is_flag=True, help='Dump raw json response from endpoint')
@pass_context
def process_node(ctx, is_raw):
    process = ctx.obj.process
    node = process['details'].get('node', {})
    if is_raw:
        echo(dumps(node))
        return
    echo("Showing node information from when the process was run...  ")
    keys = ['name', 'creation_timestamp', 'labels', 'capacity', 'node_info']
    output_pretty_json(node, keys=keys)


@process.command('events', help="Get nextflow events for this process")
@option('--raw', 'is_raw', is_flag=True, help='Dump raw json response from endpoint')
@pass_context
def process_events(ctx, is_raw):
    process = ctx.obj.process
    events = ctx.obj.session.get(process['links']['events']).json()['events']

    if is_raw:
        print(dumps(events))
        return

    headers = ['first_event_date', 'last_event_date', 'event', 'process', 'status']
    table = []
    for event in events:
        dt = dateutil.parser.parse(event['event_date'])
        try:
            last_dt = dateutil.parser.parse(event['last_event_date'])
        except Exception:
            last_dt = '-'
        row = [
            dt,
            last_dt,
            event['event'],
            (event.get('trace') or {}).get('name'),
            ((event.get('trace') or {}).get('status') or ""),
        ]
        if (event['num_events'] or 1) > 1:
            row[-1] += ' ( x%s )' % event['num_events']
        table.append(row)

    tbl = tabulate(table, headers=headers)
    echo(tbl)


@job.command(help="Cancel a running job")
@pass_context
def cancel(ctx):
    workflow_job = ctx.obj.job
    try:
        status_message = workflow_job.cancel()
    except ServerError as e:
        abort(e)
    secho("Job has been cancelled. Response was: %s" % status_message, bold=True)


@job.command(help="Monitor a running job.")
@option(
    '-o', '--output', type=click.Choice(['normal', 'wide']), default='normal', help='Format output'
)
@pass_context
def watch(ctx, output):
    workflow_job = ctx.obj.job
    try:

        status = workflow_job.status
        if not workflow_job.running:
            print_warn("Job is not in running state. Nothing to watch.")
            ctx.invoke(view)
            return

        last_status = None
        while status in INIT_STATUSES:

            status = workflow_job.status
            if status != last_status:
                sys.stdout.write("\nJob status: %s." % status)
                sys.stdout.flush()
                last_status = status

            while workflow_job.status == last_status and last_status in INIT_STATUSES:
                sys.stdout.write(".")
                sys.stdout.flush()
                time.sleep(5.0)
                workflow_job.refresh()
                if check_for_broken_job(workflow_job):
                    return

        if workflow_job.running:
            import curses

            stdscr = curses.initscr()
            curses.noecho()
            try:
                stdscr.clear()
                stdscr.addstr("\nJob %s is running." % workflow_job.job_id)
                n = 0
                while workflow_job.running:
                    stdscr.addstr(".")
                    time.sleep(1.0)
                    workflow_job.refresh()
                    processes_resource = workflow_job.processes()

                    stdscr.clear()
                    stdscr.addstr("\nJob %s is running." % workflow_job.job_id)
                    for i in range(n):
                        stdscr.addstr(".")
                    n += 1
                    if n > 3:
                        n = 0
                    stdscr.addstr("\n\n")
                    if not processes_resource:
                        stdscr.addstr("No processes running")
                    else:
                        steps_txt = get_processes_table(
                            processes_resource, is_color=False, is_wide=(output == 'wide')
                        )
                        try:
                            stdscr.addstr(steps_txt)
                        except Exception:
                            pass
                    stdscr.refresh()
                time.sleep(1.0)
            finally:
                curses.endwin()
            col = status_to_color.get(workflow_job.status, '')
            echo(
                "\nJob has completed in %s with status: %s"
                % (
                    str(workflow_job.duration).split('.')[0],
                    click.style(workflow_job.status, fg=col, bold=True),
                )
            )
            cmd = ['nextcode', 'workflow', 'job', str(workflow_job.job_id), 'view']
            echo("Executing %s..." % ' '.join(cmd))
            subprocess.call(cmd)

        else:
            print_error("\nUnexpected job status: %s" % workflow_job.status)
            echo("\nView job details with: 'nextcode workflow job %s view'" % workflow_job.job_id)
    except KeyboardInterrupt:
        echo(
            "\nStopped watching job. To continue run: "
            "nextcode workflow job %s watch" % workflow_job.job_id
        )
        return


@job.command()
@argument('link_type', required=True, type=click.Choice(['kibana', 'grafana']))
@pass_context
def link(ctx, link_type):
    """
    Helper that produces a link to external monitoring tools with filters
    preconfigured for this job
    """

    workflow_job = ctx.obj.job

    start_date = workflow_job.submit_date
    end_date = datetime.datetime.utcnow()
    if workflow_job.complete_date:
        end_date = workflow_job.complete_date
    end_date += datetime.timedelta(minutes=5)
    job_id = workflow_job.job['job_id']
    profile_name = get_default_profile()
    if link_type == 'kibana':
        kibana_url = ctx.obj.session.url_from_endpoint("kibana")

        start_date = start_date.isoformat()
        end_date = end_date.isoformat()
        link = (
            "{kibana_url}/discover?_g=(refreshInterval:(pause:!t,value:0),"
            "time:(from:'{start_date}',mode:absolute,to:'{end_date}'))"
            "&_a=(columns:!(kubernetes.labels.taskName,message),interval:auto,"
            "query:(language:lucene,query:'kubernetes.labels.job-id:{job_id}%20%26%26%20kubernetes.labels.started-by:workflow-service'),"
            "sort:!('@timestamp',desc))".format(
                kibana_url=kibana_url,
                start_date=start_date,
                end_date=end_date,
                job_id=job_id,
            )
        )
        echo(link)
        click.launch(link)
    elif link_type == 'grafana':
        parts = urlsplit(ctx.obj.service.base_url)
        grafana_url = urljoin('%s://%s' % (parts.scheme, parts.netloc), 'admin/grafana')

        from_timestamp = int(time.mktime(start_date.timetuple()) * 1000)
        to_timestamp = int(time.mktime(end_date.timetuple()) * 1000)
        link = (
            "{grafana_url}/d/KBmVXHXnas/workflow-service-job-details?orgId=1&"
            "from={from_timestamp}&to={to_timestamp}&"
            "var-job_id={job_id}&var-step=All&refresh=5s".format(
                grafana_url=grafana_url,
                from_timestamp=from_timestamp,
                to_timestamp=to_timestamp,
                job_id=job_id,
            )
        )
        echo(link)
        click.launch(link)



#!/usr/bin/env python

import click
from nextcodecli import get_version_string
from nextcodecli.utils import dumps, print_table


@click.command(help="Show the status of the pipelines-service")
@click.option('-r', '--raw', 'is_raw', is_flag=True, help='Dump raw json response')
@click.pass_context
def status(ctx, is_raw):
    svc = ctx.obj.service
    session = svc.session
    profile_config = svc.client.profile.content
    profile_name = svc.client.profile_name
    if is_raw:
        dct = {
            'version': get_version_string(),
            'profile': profile_name,
            'profile_config': profile_config,
        }
        for key in ('build_info', 'app_info', 'endpoints', 'current_user'):
            dct[key] = getattr(svc, key, None)
        resp = session.get(session.url_from_endpoint('pipelines'))
        dct['pipelines'] = [p['name'] for p in resp.json()['pipelines']]
        click.echo(dumps(dct))
    else:
        click.echo(get_version_string() + '\n')

        click.echo("{0:20}{1}".format('profile', profile_name))
        click.echo('')
        print_table(svc.build_info)
        print_table(svc.app_info)

        click.echo('')



#!/usr/bin/env python

import click
from datetime import datetime
from nextcode import Client
from nextcode.config import get_default_profile
from nextcodecli.utils import print_tab, dumps


@click.command()
@click.option(
    '-r',
    '--raw',
    'is_raw',
    is_flag=True,
    help='Return decoded token as json as well as the token itself',
)
def cli(is_raw):
    """
    Print out an access token for the current profile
    """
    client = Client()
    token = client.get_access_token(decode=False)
    payload = client.get_access_token(decode=True)
    profile = get_default_profile()
    if is_raw:
        ret = {"token": token, "payload": payload, "profile": profile}
        click.echo(dumps(ret))
        return

    click.echo(
        "Access token for services in the '%s' profile. Use by adding 'Authorization: Bearer xxx' request header"
        % profile
    )
    print_tab("Issuer", payload['iss'])
    exp = datetime.utcfromtimestamp(payload['exp'])
    diff = exp - datetime.utcnow()
    print_tab("Expires in", diff)
    click.echo('\n' + token + '\n')



#!/usr/bin/env python
import click
import sys
import collections
from tabulate import tabulate
import dateutil
import yaml
from pathlib import Path

from nextcode.exceptions import ServerError
from nextcode.services.query.exceptions import TemplateError

from ...queryapi import get_results
from ...utils import print_tab, dumps, get_logger, abort

log = get_logger(name='commands.query', level='INFO')


def fmt_date(dt):
    return dateutil.parser.parse(dt).strftime("%Y-%m-%d %H:%M")


@click.group(help="Query Template management")
def templates():
    pass


@templates.command()
@click.option('-o', '--organization', help="The organization to show templates for")
@click.option('-c', '--category', help="Category for templates")
@click.option(
    '--raw', 'is_raw', is_flag=True, default=False, help='Raw JSON response from endpoint'
)
@click.pass_context
def list(ctx, organization, category, is_raw):
    """
    Show a list of all templates registered into the system

    The list can optionally be filtered and output as json for additional
    processing.
    """
    svc = ctx.obj.service
    all_templates = svc.get_templates(organization=organization, category=category)
    template_list = []
    for name, templ in all_templates.items():
        full_name = templ['full_name']
        template_list.append(
            [
                templ['id'],
                templ['name'],
                templ['version'],
                full_name,
                fmt_date(templ['create_date']),
                templ['list_description'],
            ]
        )

    tbl = tabulate(
        template_list, headers=['id', 'name', 'version', 'full_name', 'date', 'description']
    )
    click.echo(tbl)


@templates.command()
@click.argument('name', nargs=1)
@click.pass_context
def delete(ctx, name):
    """
    Delete a template by full name (e.g. [org]/[category]/[name]/[version])
    """
    svc = ctx.obj.service
    try:
        svc.delete_template(name)
    except TemplateError as ex:
        abort(ex)
    click.secho("Template '%s' has been deleted." % (name), bold=True)


@templates.command()
@click.argument('name', nargs=1)
@click.option(
    '--raw', 'is_raw', is_flag=True, default=False, help='Raw JSON response from endpoint'
)
@click.option('--yaml', 'is_yaml', is_flag=True, default=False, help='Raw yaml template')
@click.option('--args', 'is_args', is_flag=True, default=False, help='View arguments')
@click.pass_context
def view(ctx, name, is_raw, is_yaml, is_args):
    """
    View details about a single template

    Use the full name of the template including organization, 
    category and version, i.e. [org]/[cat]/[name]/[version]

    If you omit the version the latest version is fetched.
    e.g. 'nextcode query templates view wxnc/system/acmg_scores'

    You can also specify a glob pattern. e.g. 
    'nextcode query templates view "wxnc/system/acmg_scores/1.4.*"'
    """
    svc = ctx.obj.service
    try:
        template = svc.get_template(name)
    except ServerError:
        abort("Template {} not found".format(name))

    if is_yaml:
        click.echo(template["file_contents"])
        return

    if is_args:
        if is_raw:
            click.echo(dumps(template['arguments']))
            return
        args_list = []
        for arg in template['arguments']:
            args_list.append(
                [
                    arg['name'],
                    arg.get('type'),
                    arg.get('optional'),
                    ", ".join(arg.get('values', [])),
                    arg.get('default', "")[:50],
                ]
            )
        args_list.sort()
        tbl = tabulate(args_list, headers=['name', 'type', 'optional', 'values', 'default'])
        click.echo(tbl)
        return
    if is_raw:
        click.echo(dumps(template))
        return
    print_tab('Template ID', template['id'])
    print_tab('Name', template['name'])
    print_tab('Organization', template['organization'])
    print_tab('Category', template['category'])
    print_tab('Date Created', fmt_date(template['create_date']))
    print_tab('Version', template['version'])
    print_tab('Description', template['list_description'])
    args = []
    for arg in template['arguments']:
        if not arg.get('optional'):
            args.append(arg['name'])
    print_tab('Required Arguments', ", ".join(args))
    print_tab('Perspectives', ", ".join([p['name'] for p in template['perspectives']]))


@templates.command()
@click.argument('filename', nargs=1)
@click.option('--replace', is_flag=True, default=False, help='Delete and replace existing template')
@click.pass_context
def add(ctx, filename, replace):
    """Add a new template from yaml file.
    """
    svc = ctx.obj.service
    try:
        full_name = svc.add_template_from_file(filename, replace=replace)
    except TemplateError as ex:
        abort(ex)

    click.secho("Template '%s' has been successfully added" % (full_name), bold=True)
    click.echo("View the template with: nextcode query templates view {}".format(full_name))


def _render(svc, name, args):
    try:
        _ = svc.get_template(name)
    except TemplateError as ex:
        abort(str(ex))
    params = {}
    for a in args:
        lst = a.split("=", 1)
        try:
            params[lst[0]] = lst[1]
        except IndexError:
            abort("Arguments invalid. Should be 'args=val")
    try:
        query_string = svc.render_template(name, params)
    except TemplateError as ex:
        if "Missing arguments" in str(ex):
            click.secho(str(ex), fg='red')
            click.echo(
                "Hint: You can view arguments for this template with: nextcode query templates view %s --args"
                % name
            )
            sys.exit(1)
        else:
            abort(ex)
    return query_string, params


@templates.command(context_settings=dict(ignore_unknown_options=True))
@click.argument('name', nargs=1)
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def render(ctx, name, args):
    """
    Render a query from template

    View the query that will be generated from the template with
    the specified parameters.

    Use the full name of the template in the form
    [org]/[cat]/[name]/[version (optional)]

    Example:
    'nextcode query templates render wxnc/category/template/1.0.0 arg1=1 arg2=2'

    """
    svc = ctx.obj.service
    query_string, _ = _render(svc, name, args)

    click.echo(query_string)


@templates.command(context_settings=dict(ignore_unknown_options=True))
@click.argument('name', nargs=1)
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
@click.option('-o', '--output', 'output_file', help='File to write results to (otherwise output to console)')
@click.option('-l', '--limit', 'limit', default=0, help='Limit the number of rows returned', show_default=True)
@click.option(
    '-n',
    '--nowait',
    is_flag=True,
    default=False,
    help='Run command without waiting for it to finish',
)
@click.option(
    '--raw', 'is_raw', is_flag=True, default=False, help='Raw JSON response from endpoint'
)
@click.pass_context
def execute(ctx, name, args, output_file, limit, nowait, is_raw):
    """
    Execute a template

    Execute a query that will be generated from the template with
    the specified parameters.

    Use the full name of the template in the form
    [org]/[cat]/[name]/[version (optional)]

    Example:
    'nextcode query templates execute wxnc/category/template/1.0.0 arg1=1 arg2=2'

    """
    svc = ctx.obj.service
    # render the template to ensure it is okay.
    query_string, params = _render(svc, name, args)

    qry = svc.execute_template(name, **params, nowait=True)
    if not nowait:
        qry.wait()

    if qry.failed:
        click.secho(
            "Query {} failed with message: {}".format(qry.query_id, qry.status_message), fg="red"
        )
        return
    if is_raw:
        click.echo(dumps(qry.raw))
        return

    if qry.running:
        click.echo("Query {} is now running.".format(qry.query_id))
        click.echo("View query process with: nextcode query info {}".format(qry.query_id))
        return
    click.secho(
        "Query {} completed successfully in {} ms".format(qry.query_id, qry.duration), fg="green"
    )

    get_results(qry, limit, None, None, output_file)



#!/usr/bin/env python

import click
import dateutil
import json
from tabulate import tabulate
from textwrap import dedent
from nextcodecli.utils import dumps, abort
from nextcodecli.pipelines.cmdutils import status_to_color


@click.command(help="List running jobs. Use switches to filter.")
@click.option('--mine', 'is_mine', is_flag=True, help='List jobs from me')
@click.option('--user', 'user_name', help='User to filter for')
@click.option('--all', 'is_all', is_flag=True, help='Do not only list running jobs')
@click.option('--latest', 'is_latest', is_flag=True, help='View the top 10 latest submitted jobs')
@click.option('--raw', 'is_raw', is_flag=True, help='Dump raw json for further processing')
@click.option('--page', 'is_page', is_flag=True, help='Page results')
@click.option('--status', 'status', default=None, help='Filter status')
@click.option('--start-date', 'start_date', default=None, help='Only show jobs after this date')
@click.option('--end-date', 'end_date', default=None, help='Only show jobs before (and not including) this date')
@click.option('--num', default=50, help='Maximum number of jobs to return')
@click.option(
    '--filter',
    'job_filter',
    default='',
    help='Filter for jobs, with format KEY=VALUE,' 'KEY2=VALUE2',
)
@click.option(
    '--fields',
    'add_fields',
    default='',
    help='Additional fields you wish to view, '
    'separated by comma. Also supports '
    'removing default fields by adding a - '
    'to the beginning of the fieldname.',
)
@click.pass_context
def jobs(
    ctx, is_mine, user_name, is_all, is_latest, is_raw, is_page, status, start_date, end_date, num, job_filter, add_fields
):

    svc = ctx.obj.service
    session = svc.session
    fields = [
        'job_id',
        'user_name',
        'date_submitted',
        'duration',
        'pipeline_name',
        'sample_name',
        'region',
        'status',
    ]

    rm_missing = []
    if add_fields:
        try:
            for f in add_fields.split(','):
                if f[0] == '-':
                    try:
                        fields.remove(f[1:])
                    except ValueError:
                        rm_missing.append(f[1:])
                elif f not in fields:
                    fields.append(f)
            if rm_missing:
                rm_err = "Error: Field(s) {} not in default fields, ignored.".format(
                    ','.join(sorted(rm_missing))
                )
                click.echo(click.style(rm_err, fg='red'))
        except Exception:
            abort("Fields are incorrectly formatted")
    data = {'field': fields, 'start_date': start_date, 'end_date': end_date}
    if status:
        data['status'] = status.upper()
        if is_mine or is_latest or is_all:
            abort("You cannot combine --status with other filter flags")

    if is_mine:
        data['user_name'] = session.current_user['preferred_username']
        data['status'] = 'ALL'
    elif user_name:
        data['user_name'] = user_name
        data['status'] = 'ALL'
    elif is_latest or is_all:
        data['status'] = 'ALL'
    data['limit'] = num
    if job_filter:
        try:
            data.update({i.split('=')[0]: i.split('=')[1] for i in job_filter.split(',')})
        except Exception as e:
            abort("Invalid filter format: KEY=VALUE expected, separated by comma.")
    resp = session.get(session.url_from_endpoint('jobs'), json=data)
    try:
        resp.raise_for_status()
    except Exception as e:
        abort(json.dumps(resp.json(), indent=4))
    jobs = resp.json()['jobs']
    if is_raw:
        click.echo(dumps(jobs))
        return

    if is_mine:
        click.echo("All my jobs:")
    elif is_latest:
        click.echo("Latest jobs:")
    elif is_all:
        click.echo("All jobs:")
    else:
        click.echo("Currently running jobs:")
    if jobs:
        err = set()
        jobs_list = []
        for j in jobs:
            job_list = []
            det = j.get('details', {})
            if det and isinstance(det, dict):
                j = dict(list(det.items()) + list(j.items()))
                j.pop('details')
            date_submitted = None
            for f in fields:
                try:
                    v = j[f]
                    if isinstance(v, dict):
                        v = dedent(json.dumps(v, indent=4, separators=(',', ':'))[1:-1])
                    elif isinstance(v, list):
                        v = '\n'.join(
                            dedent(json.dumps(h, indent=4, separators=(',', ':'))[1:-1]) for h in v
                        )
                    elif f == 'date_submitted':
                        date_submitted = dateutil.parser.parse(v)
                        v = dateutil.parser.parse(v).strftime("%Y-%m-%d %H:%M")
                    elif f == 'duration' and v:
                        hours, remainder = divmod(v, 3600)
                        minutes, seconds = divmod(remainder, 60)
                        v = '%02d:%02d:%02d' % (int(hours), int(minutes), int(seconds))
                    elif f == 'status':
                        col = status_to_color.get(v, '')
                        v = click.style(v, fg=col, bold=True)
                    elif f == 'spot':
                        v = "Yes" if v else "No"
                        spot_instance = (j.get('instances') or [{}])[0].get("is_spot")
                        if v == "Yes" and not spot_instance:
                            v += click.style(" NO", fg='red')
                    job_list.append(v)
                except KeyError:
                    err.add(f)
            if det.get("old_job_id"):
                job_list[0] = '{} ({})'.format(j['job_id'], det.get("old_job_id"))
            jobs_list.append(job_list)
        [fields.remove(j) for j in err]
        if is_latest:
            jobs_list = jobs_list[:10]
        tbl = tabulate(jobs_list, headers=fields)
        if len(jobs_list) > 30 and is_page:
            click.echo_via_pager(tbl)
        else:
            click.echo(tbl)
            click.echo("{} jobs found".format(len(jobs_list)))
        if err:
            out = "Error: Field(s) {} not found, skipped.".format(','.join(sorted(err)))
            click.echo(click.style(out, fg='red'))
    else:
        click.echo(
            "No jobs found. Run 'nextcode pipelines jobs --latest' "
            "to view the latest completed jobs"
        )



#!/usr/bin/env python

import click

from functools import wraps

from nextcode import get_service
from nextcode.exceptions import ServerError, InvalidProfile
from nextcodecli.utils import print_error, abort, check_profile

status_to_color = {
    'PENDING': 'yellow',
    'COMPLETED': 'green',
    'STARTED': 'yellow',
    'ERROR': 'red',
    'CANCELLED': 'white',
}


def check_project(func):
    @wraps(func)
    def _check(*args, **kwargs):
        ctx = click.get_current_context()
        svc = ctx.obj.service
        if not svc.project:
            abort("Please set a project by running: nextcode query project [name]")
        else:
            return func(*args, **kwargs)

    return _check


@click.group()
@click.pass_context
def cli(ctx):
    """Root subcommand for query api functionality.

    This command allows you to run GOR queries against the selected server.
    Before running a query you must select a project with 'nextcode query project [name]'.

    You must have access to the project in CSA in order to run queries. If needed you can
    user the 'nextcode csa' command to add yourself to a project, provided you have a CSA
    admin username and password to give yourself the roles."""
    if ctx.obj.service is None:
        check_profile(ctx)
        ctx.obj.service = ctx.obj.client.service("query")


from nextcodecli.commands.query.run import run, project, info, results, cancel, list_queries, progress, details
from nextcodecli.commands.query.templates import templates
from nextcodecli.commands.query.status import status

cli.add_command(run)
cli.add_command(project)
cli.add_command(info)
cli.add_command(results)
cli.add_command(cancel)
cli.add_command(progress)
cli.add_command(list_queries)
cli.add_command(details)

cli.add_command(templates)

cli.add_command(status)



#!/usr/bin/env python

import os
import sys
import pprint
import nextcode
from nextcodecli.utils import abort
from click import echo, secho

WARN_MSG = "Mismatched %d columns in line %d, but %s in header"


class TsvDict(dict):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def get_item(self, key, default):
        value = self.get(key, default)
        if value == '':
            value = default

        return value


def parse(file_name):
    with open(file_name, 'r') as f:
        lines = f.readlines()

    samples = {}

    if len(lines) == 0:
        return samples

    headerline = lines[0]

    # Get an array of all the column names in lowercase to avoid
    # case sensitivity issues
    colnames = headerline.rstrip("\n").lower().split("\t")
    for line_num, line in enumerate(lines[1:]):
        parts = line.rstrip("\n").split("\t")
        dct = TsvDict()
        if len(parts) != len(colnames):
            exit_msg = WARN_MSG % (len(parts), line_num + 2, len(colnames))
            raise ValueError(exit_msg)
        else:
            dct = TsvDict(zip(colnames, parts))
            if dct['#sampleid'] not in samples:
                sample = {
                    'subject_id': dct.get_item('subjectid', None),
                    'gender': dct.get_item('gender', None),
                    'product': dct.get_item('product', None),
                    'product_version': dct.get_item('productversion', None),
                    'study_id': dct.get_item('studyid', None),
                    'study_role': dct.get_item('studyrole', None),
                    'affected': dct.get_item('affected', None),
                    'sample_tags': dct.get_item('sampletags', None),
                    'files': [],
                }
                samples[dct['#sampleid']] = sample

            file = {
                'platform': dct.get_item('platform', None),
                'file_type': dct.get_item('filetype', None),
                'compression': dct.get_item('compression', None),
                'data_type': dct.get_item('datatype', None),
                'read_no': dct.get_item('readno', None),
                'part_no': dct.get_item('partno', None),
                'file_name': dct.get_item('filename', None),
                'path': dct.get_item('path', None),
                'md5': dct.get_item('md5', None),
                'file_tags': dct.get_item('filetags', {}),
            }

            samples[dct['#sampleid']]['files'].append(file)

    return samples


def is_remote(path):
    parts = path.split('://')
    return len(parts) > 1


def create_sample_data(sample, sample_id, subject_id, org_key, project_name, skip_processing_state=False):
    sample_data = {
        'extid': sample_id,
        'sample_data_type': sample.get('product'),
        'sample_data_type_version': sample.get('product_version'),
        'initial_study': {
            'extid': sample.get('study_id'),
            'participant_kind': sample.get('study_role'),
            'affected': sample.get('affected'),
        },
        'subject': {
            'extid': subject_id,
            'gender': sample.get('gender'),
            'organization_key': org_key,
            'initial_project_key': project_name,
        },
        'tags': sample.get('sampletags', {}),
        'sample_data_files': [],
    }

    if not skip_processing_state:
        sample_data['processing_state'] = 'uploaded'
    return sample_data


def create_file_data(file):
    file_path = file['path']
    if is_remote(file_path):
        full_path = file_path
    else:
        full_path = os.path.abspath(os.path.expanduser(file_path))

    file_tags = file.get('file_tags', {})
    tags = dict(
        {
            'filetype': file['file_type'],
            'platform': file['platform'],
            'datatype': file['data_type'],
            'read_no': file['read_no'],
            'part_no': file['part_no'],
        },
        **file_tags
    )

    file_data = {
        'file_name': file['file_name'],
        'full_path': full_path,
        'platform': file['platform'],
        'sample_type': file['data_type'],
        'file_type': file['file_type'],
        'read_no': file['read_no'],
        'part_no': file['part_no'],
        'md5': file['md5'],
        'tags': tags,
    }
    return file_data


def process_sample(session, params):
    sample = params.get('sample', None)
    org_key = params.get('org_key', None)
    sample_id = params.get('sample_id', None)
    subject_id = params.get('subject_id', None)
    project_name = params.get('project_name', None)
    skip_processing_state = params.get('skip_processing_state', False)
    sample_data = create_sample_data(
        sample, sample_id, subject_id, org_key, project_name, skip_processing_state
    )

    for file in sample.get('files'):
        file_data = create_file_data(file)
        sample_data['sample_data_files'].append(file_data)

    try:
        result = session.import_sample(sample_data)
        echo('%s imported successfully to %s' % (sample_id, project_name))
    except nextcode.exceptions.CSAError as ex:
        secho(f"Error occurred when importing sample '{sample_id}'", fg="red")
        abort(ex)
    return result



#!/usr/bin/env python

import os
import click
import requests
import subprocess
import hjson

from nextcode.packagelocal import package_and_upload
from nextcodecli.pipelines.jobs import post_job
from nextcodecli.utils import dumps, abort, print_warn, print_error


@click.command(help="Run a pipeline job. Pass in a manifest file for the job definition.")
@click.argument('job_filename', required=True)
@click.option('--local', 'local_folders', help='Run your local etl-pipelines folder in the worker')
@click.option(
    '--git',
    'git_branch',
    help='Run a specific git branch (or tag or sha), overriding the job manifest',
)
@click.option(
    '--pipeline-definition', help='Override the service pipeline definition with a local file'
)
@click.option(
    '--skip-definition',
    is_flag=True,
    default=False,
    help='Skip the pipeline definition validation.',
)
@click.option(
    '-f',
    '--force-start',
    is_flag=True,
    default=False,
    help='Start the worker immediately instead of going through the queue',
)
@click.option(
    '--test',
    'is_test',
    is_flag=True,
    default=False,
    help='Do not actually submit the job but just test basic validation',
)
@click.option('-q', '--nowait', help='Do not execute the watch command for a running pipeline')
@click.pass_context
def run(
    ctx,
    job_filename,
    local_folders,
    git_branch,
    pipeline_definition,
    skip_definition=False,
    force_start=False,
    is_test=False,
    nowait=False,
):
    svc = ctx.obj.service
    session = ctx.obj.session
    if not os.path.exists(job_filename):
        abort("Job filename '%s' not found" % job_filename)

    pipeline_definition_contents = None
    if pipeline_definition:
        path = os.path.expanduser(pipeline_definition)
        print_warn("Uploading local definition from {}".format(path))
        try:
            with open(path, 'r') as f:
                pipeline_definition_contents = hjson.load(f)
        except Exception as e:
            print_error(e)
            abort("Error loading pipeline definition: %s" % pipeline_definition)
    if skip_definition:
        print_warn("Skipping pipeline definition")
        pipeline_definition_contents = {}

    build_source = None
    build_context = None

    if local_folders:
        local_packages = []
        for local_folder in local_folders.split(','):
            project_path = os.path.expanduser(local_folder)
            if project_path.endswith('/'):
                project_path = project_path[:-1]
            project_path = os.path.abspath(project_path)

            project_name = 'etl-pipelines'

            # ! special case for duct tools upload
            if 'duct-tools' in project_path:
                project_name = project_path.split('/')[-1]

            p = package_and_upload(svc, project_name, project_path)
            local_packages.append(p)

        build_source = 'url'
        build_context = ','.join(local_packages)
    elif git_branch:
        build_source = 'git'
        build_context = git_branch

    try:
        manifest = None
        try:
            with open(job_filename, 'r') as f:
                manifest = hjson.load(f)
        except Exception as e:
            raise RuntimeError("Manifest file '%s' is invalid: %s" % (job_filename, e))

        responses = post_job(
            session,
            manifest,
            build_source,
            build_context,
            pipeline_definition_contents,
            force_start=force_start,
            is_test=is_test,
        )
    except Exception as e:
        print_error('ERROR posting job: %s' % str(e))
        raise
    # pylint: disable=E1101
    for resp in responses:
        if resp.status_code not in (requests.codes.ok, requests.codes.created):
            try:
                print_error(
                    "Error creating job: %s"
                    % (resp.json()['error']['description'] or resp.json()['message'])
                )
            except Exception:
                print_error(resp.text)
        else:
            click.echo(
                "Job for sample {} submitted. View job with: "
                "nextcode pipelines job {} view".format(
                    resp.json().get('sample_name', '').encode('utf-8'), resp.json().get('job_id')
                )
            )

    if is_test:
        try:
            formatted_json = dumps(resp.json())
            click.echo("Response:")
            click.echo(formatted_json)
        except Exception:
            click.echo(resp.text)
        click.secho("This was just a test. No job was actually submitted", fg='yellow', bold=True)
        return

    if not nowait:
        # typical case of single job
        if len(responses) == 1 and 200 <= responses[0].status_code <= 201:
            job_id = resp.json()['job_id']
            cmd = ['nextcode', 'pipelines', 'job', str(job_id), 'watch']
            click.echo("Executing %s..." % ' '.join(cmd))
            subprocess.call(cmd)



#!/usr/bin/env python
import os
import click
from click import echo, secho, pass_context, command, argument, option

import hjson
import logging
import subprocess

from nextcode.packagelocal import package_and_upload
from nextcode.exceptions import ServerError, UploadError

from nextcodecli.utils import abort

log = logging.getLogger(__name__)


@command()
@argument('pipeline_name', required=False)
@argument('project_name', required=False)
@option('-p', '--parameters', help="Parameters for the run", multiple=True)
@option('-s', '--script', help="Git link to a workflow (only for custom)")
@option('-r', '--revision', help="Git SHA to run")
@option('--profile', help="Nextflow configuration profile to use")
@option('--local', 'local_folder', help='Run a local nextflow folder')
@option('-f', '--file', 'job_filename', help='Run a manifest file')
@option(
    '-t',
    '--trace',
    is_flag=True,
    help='Injece trace directives to get more debugging information in nextflow log',
)
@option(
    '-q', '--nowait', is_flag=True, help='Do not execute the watch command for a running pipeline'
)
@option('-d', '--description', help='A human readable description of the job')
@pass_context
def run(
    ctx,
    pipeline_name,
    project_name,
    parameters,
    script,
    revision,
    profile,
    local_folder,
    job_filename,
    nowait,
    trace,
    description,
):
    """
    Start a new nextflow job.

    Specify a named workflow and project or a manifest via --file.

    You can override all parameters in the file on the command-line via '-p key1=val1 -p key2=val2'

    If you are running against a debug server that allows direct script execution you
    can use 'custom' for pipeline_name.
    """
    manifest = {}
    params = {}
    if job_filename:
        with open(job_filename, 'r') as f:
            manifest = hjson.load(f)
        pipeline_name = pipeline_name or manifest.get('pipeline_name')
        project_name = project_name or manifest.get('project_name')
        script = script or manifest.get('script')
        revision = revision or manifest.get('revision')
        profile = profile or manifest.get('profile')
        params = manifest.get('params', {})
    try:
        override_params = {}
        for p in parameters:
            lst = p.split("=", 1)
            override_params[lst[0]] = lst[1].strip()
    except Exception:
        abort("Override parameters are badly formed. Expected '-p k1=v1 -p k2=v2'")

    for k, v in override_params.items():
        params[k] = v

    log.info(
        "Running pipeline '%s' on project '%s' with parameters %s",
        pipeline_name,
        project_name,
        dict(params),
    )

    if not project_name:
        abort("You must specify PROJECT_NAME")
    if pipeline_name == 'custom' or not pipeline_name:
        if not script and not local_folder:
            abort(
                "If you are executing a custom script the you must specify a script or local folder"
            )
        pipeline_name = None
        if script and local_folder:
            abort("You cannot specify both a script and local folder")

    elif script or local_folder:
        abort("You cannot specify a script or local folder if you are running a named workflow")

    build_source = 'builtin'
    build_context = pipeline_name
    if local_folder:
        path = os.path.abspath(os.path.expanduser(local_folder))
        try:
            p = package_and_upload(ctx.obj.service, 'local_workflow', path)
        except UploadError as ex:
            abort(ex)
        build_source = 'url'
        build_context = p
    elif script:
        build_source = 'git'
        build_context = script

    try:
        job = ctx.obj.service.post_job(
            pipeline_name,
            project_name,
            params,
            script,
            revision or None,
            build_source,
            build_context,
            profile,
            trace,
            description=description,
        )
    except ServerError as e:
        abort(str(e))

    secho('Job %s has been submitted' % job.job_id, bold=True)

    if not nowait:
        job_id = job.job_id
        cmd = ['nextcode', 'workflow', 'job', str(job_id), 'watch']
        echo("Executing %s..." % ' '.join(cmd))
        subprocess.call(cmd)



import sys
import click
from nextcode import get_service
from nextcode.exceptions import InvalidProfile
from nextcodecli.utils import check_profile

status_to_color = {
    'PENDING': 'yellow',
    'COMPLETED': 'green',
    'STARTED': 'yellow',
    'ERROR': 'red',
    'KILLED': 'red',
    'CANCELLED': 'white',
}


@click.group()
@click.pass_context
def cli(ctx):
    """Root subcommand for workflow functionality"""
    if ctx.obj.service is None:
        check_profile(ctx)
        ctx.obj.service = ctx.obj.client.service("workflow")
        ctx.obj.session = ctx.obj.service.session
        ctx.obj.job = None


from nextcodecli.commands.workflow.jobs import jobs
from nextcodecli.commands.workflow.job import job
from nextcodecli.commands.workflow.run import run
from nextcodecli.commands.workflow.projects import projects
from nextcodecli.commands.workflow.status import status, smoketest
from nextcodecli.commands.workflow.pipelines import pipelines

cli.add_command(jobs)
cli.add_command(job)
cli.add_command(run)
cli.add_command(projects)
cli.add_command(status)
cli.add_command(smoketest)
cli.add_command(pipelines)



#!/usr/bin/env python

import sys
import click
import time
import dateutil
import datetime
from tabulate import tabulate
import requests

from nextcodecli.pipelines.jobs import find_job, download_output_file
from nextcodecli.pipelines.cmdutils import get_steps_table, print_filelist, status_to_color
from nextcodecli.utils import dumps, abort, print_tab, print_warn, print_error

INIT_STATUSES = ('QUEUED', 'INITIALIZING', 'LAUNCHING')


def has_failed_steps(job_resource):
    for step in job_resource.get('steps', []):
        if step['status'] == 'FAILED':
            return True
    return False


@click.command(
    cls=click.Group,
    invoke_without_command=True,
    help="View or manage individual jobs. Call with [job_id] or latest to "
    "view your latest submitted job",
)
@click.argument('job_id', required=True)
@click.pass_context
def job(ctx, job_id):
    if job_id != 'latest':
        try:
            job_id = int(job_id)
        except ValueError:
            abort("job_id should be a number or 'latest', not '%s'" % job_id)
    job_resource = find_job(ctx.obj.session, job_id)
    if not job_resource:
        abort("Job not found")

    ctx.obj.job = job_resource
    ctx.obj.job_id = job_id
    if ctx.invoked_subcommand is None:
        ctx.invoke(view)
        click.echo(job.get_help(ctx))


@job.command(help="View information about the job")
@click.option('--raw', 'is_raw', is_flag=True, help='Dump raw json response from endpoint')
@click.pass_context
def view(ctx, is_raw):
    job_resource = ctx.obj.job
    session = ctx.obj.session
    if is_raw:
        click.echo(dumps(job_resource))
        return
    else:
        status = job_resource['status']
        for k in (
            'job_id',
            'date_submitted',
            'date_completed',
            'pipeline_name',
            'sample_name',
            'description',
        ):
            print_tab(k, job_resource[k])

        mr = ""
        if job_resource['details'].get('submitter'):
            if job_resource['details']['submitter']['user_name'] != job_resource['user_name']:
                mr = " (submitted by %s)" % job_resource['details']['submitter']['user_name']
        print_tab('user_name', job_resource['user_name'] + mr)
        instance_id = '-'
        if len(job_resource['instances']) > 0:
            instance = job_resource['instances'][0]
            instance_id = instance['instance_id']
            if instance['is_spot']:
                instance_id += " (spot)"
        print_tab('instance_id', instance_id)
        print_tab('region', job_resource['region'])
        print_tab('output location', job_resource['output_config']['output_path'])
        details = job_resource['details']
        if details.get('build_source', 'ami') != 'ami':
            print_tab(
                'build source',
                "%s | %s" % (details.get('build_source'), details.get('build_context')),
            )
        if job_resource['details'].get('old_job_id'):
            print_tab('Rerun of Job', details.get('old_job_id'))

        col = status_to_color.get(status, '')
        if details.get('new_job_id'):
            status += " (rerun in job %s)" % job_resource['details'].get('new_job_id')

        print_tab('status', click.style(str(status), fg=col, bold=True))
        if status != 'FINISHED':
            click.echo('')
            click.echo(get_steps_table(job_resource))
            if has_failed_steps(job_resource):
                click.secho(
                    "\nJob has failed steps. You can analyze this with the 'steps' " "subcommand",
                    fg='red',
                )
        else:
            duration = dateutil.parser.parse(
                job_resource['date_completed']
            ) - dateutil.parser.parse(job_resource['date_submitted'])
            duration = str(duration).split('.')[0]
            print_tab('duration', duration)
            num_uncompleted = len(
                [s for s in job_resource.get('steps', []) if s['status'] != 'COMPLETED']
            )
            if num_uncompleted:
                print_warn(
                    "\nJob has %s/%s uncompleted steps. This might indicate that "
                    "there was a problem with the job."
                    % (num_uncompleted, len(job_resource['steps']))
                )
                click.echo("Please investigate using the 'steps' and 'logs' subcommands")
        messages = job_resource.get('messages')
        if messages:
            click.echo("\nMessages:")
            table = []
            for message in messages:
                table.append((message['event'].get('step_name'), message['event'].get('text')))
            tbl = tabulate(table, headers=['step name', 'message'])
            click.echo(tbl)

        click.echo("\nInput files:")
        print_filelist(job_resource.get('input'))
        if job_resource.get('output'):
            click.echo("\nOutput files:")
            print_filelist(job_resource.get('output'))

    click.echo('')
    job_id = job_resource['job_id']
    if job_resource['status'] == 'FAILED' and not has_failed_steps(job_resource):
        click.secho(
            "\nJob failed but no steps failed. "
            "Please analyze logs with 'nextcode pipelines job %s logs userdata'" % job_id,
            fg='red',
        )
        click.echo(
            "Hint: You might find the cause with one of these commands\n"
            "nextcode pipelines job {job_id} logs userdata | grep -C 5 \"Caused by\"\n"
            "nextcode pipelines job {job_id} logs userdata | grep -C 5 \"ERROR\"\n".format(
                job_id=job_id
            )
        )

    if job_resource['status'] not in INIT_STATUSES:
        click.echo("\nAvailable logs (via logs subcommand argument):")
        url = session.links(job_resource)['logs']
        resp = session.get(url)
        resp.raise_for_status()
        log_urls = resp.json()
        for url in session.links(log_urls):
            click.echo(url)
        click.echo('')
    elif job_resource['status'] == 'QUEUED':
        print_warn(
            "Job is queued. "
            "You can force-start it with 'nextcode pipelines job %s launch'" % job_id
        )
        seconds_queued = (
            datetime.datetime.utcnow() - dateutil.parser.parse(job_resource['submit_date'])
        ).total_seconds()
        if seconds_queued > 120:
            click.secho(
                "Job has been queued for %.0f minutes. You might be able to see why with "
                "'nextcode pipelines job %s events'\n" % (seconds_queued / 60, job_id),
                bold=True,
            )


@job.command(help="View job messages")
@click.option('--raw', 'is_raw', is_flag=True, help='Dump raw json response from endpoint')
@click.pass_context
def messages(ctx, is_raw):
    job_resource = ctx.obj.job
    session = ctx.obj.session
    messages = job_resource.get('messages') or []
    if is_raw:
        click.echo(dumps(messages))
        return
    if not messages:
        print_warn("No messages found")
        return
    click.echo("Messages from job %s:" % job_resource['job_id'])
    table = []
    for message in messages:
        dt = dateutil.parser.parse(message['date_time'])
        table.append(
            (
                dt.strftime("%H:%M:%S"),
                message['event'].get('step_name'),
                message['event'].get('text'),
            )
        )
    tbl = tabulate(table, headers=['timestamp', 'step name', 'message'])
    click.echo(tbl + '\n')


@job.command(
    help="View the submission manifest for the job. Note that this is not the same file as a local "
    "job.json file that was used to submit the job via the cli."
)
@click.option('--raw', 'is_raw', is_flag=True, help='Dump raw json response from endpoint')
@click.option(
    '--generate', is_flag=True, help='Attempt to generate a job.json file that the CLI can use'
)
@click.pass_context
def manifest(ctx, is_raw, generate):
    job_resource = ctx.obj.job
    details = job_resource['details']
    if generate:
        keys_to_remove = ['submitter', 'files', 'sample_name']
        details["samples"] = [{"files": details["files"], "name": details["sample_name"]}]
        details["iam_profiles"] = {"download": "", "upload": ""}
        for k in keys_to_remove:
            if k in details:
                del details[k]
        for k, v in details.copy().items():
            if v is None:
                del details[k]
        click.echo(dumps(details))
    elif is_raw:
        click.echo(dumps(details))
    else:
        for k in sorted(details):
            v = details[k]
            if isinstance(v, dict) or isinstance(v, list):
                v = dumps(v)
                for l in v.split('\n'):
                    print_tab(k, l)
                    if k:
                        k = ''

            else:
                print_tab(k, v)


@job.command(help="Cancel a running job.")
@click.pass_context
def cancel(ctx):
    session = ctx.obj.session
    job_resource = ctx.obj.job
    job_id = job_resource['job_id']
    jobs = session.get(
        session.url_from_endpoint('jobs'), json={'job_id': job_id, 'field': []}
    ).json()['jobs']
    if not jobs:
        abort("Job '%s' not found" % job_id)
    job_url = session.links(jobs[0])['self']
    resp = session.delete(job_url)
    # pylint: disable=E1101
    if resp.status_code == requests.codes.ok:
        print_warn("Job %s has been cancelled" % job_id)
        if resp.json().get('error'):
            print_error(resp.json()['error'])
    elif resp.status_code == requests.codes.bad_request:
        print_warn("Error cancelling job %s" % job_id)
        print_error(resp.json()['error']['description'])
        return
    resp.raise_for_status()


@job.command(help="Force a queued job to launch.")
@click.option('-s', '--spot', 'is_spot', is_flag=True, help='Force a spot instance')
@click.option('-d', '--dedicated', 'is_dedicated', is_flag=True, help='Force a dedicated instance')
@click.pass_context
def launch(ctx, is_spot, is_dedicated):
    session = ctx.obj.session
    job_resource = ctx.obj.job
    if job_resource['status'] != 'QUEUED':
        abort("Job is in state '%s' which is not 'QUEUED'" % job_resource['status'])
    if len(job_resource['instances']) > 0:
        abort("Job already has worker instances assigned")
    instances_url = job_resource['links']['instances']
    data = {}
    if is_spot:
        data = {'spot': True}
    elif is_dedicated:
        data = {'dedicated': True}
    resp = session.post(instances_url, json=data)
    if resp.status_code == 404:
        print_error(
            "Could not launch instance. This probably means that there are too many"
            " jobs running. If the problem persists please seek help."
        )
        print_error("Error from server: %s" % resp.json()['error']['description'])
        return

    resp.raise_for_status()
    print_warn("Job worker is launching...")


@job.command(help="View input and output files with file size information")
@click.option('-r', '--raw', 'is_raw', is_flag=True, help='Dump raw json response')
@click.pass_context
def files(ctx, is_raw):
    session = ctx.obj.session
    job_resource = ctx.obj.job
    files = session.get(session.links(job_resource)['files']).json()

    if is_raw:
        print(dumps(files))
    else:
        click.echo("\nInput files:")
        print_filelist(files.get('input'), details=True)
        if files.get('output'):
            click.echo("\nOutput files:")
            print_filelist(files.get('output'), details=True)


@job.command(help="Try to find error logs generated by the job")
@click.pass_context
def errors(ctx):
    session = ctx.obj.session
    job_resource = ctx.obj.job
    log_urls = session.links(session.get(session.links(job_resource)['logs']).json())
    print_warn("Searching for errors...")
    log_group = 'pipelines_userdata'
    filters_to_try = ['Command error', 'Caused by', 'ERROR']
    logs = ""
    for log_filter in filters_to_try:
        url = log_urls[log_group]
        if log_filter:
            url += '?filter=%s' % log_filter
        resp = session.get(url)
        if resp.status_code != 200:
            print_error(resp.json()['error']['description'])
            return
        logs = resp.text
        if logs:
            break
    if logs:
        click.echo(logs)
        click.echo("\n%s loglines found" % (logs.count('\n')))
    else:
        print_error(
            "No identifiable error logs found. You can keep searching with "
            "'nextcode pipelines job %s logs userdata'" % job_resource['job_id']
        )


@job.command(help="View logs associated with a job. Choose the log group to stream.")
@click.argument('log_group', default='')
@click.option('--filter', 'log_filter', help='Filter logs on text')
@click.pass_context
def logs(ctx, log_group, log_filter):
    session = ctx.obj.session
    job_resource = ctx.obj.job

    log_urls = session.links(session.get(session.links(job_resource)['logs']).json())
    if log_group.lower().startswith('error'):
        print_warn("Searching for errors...")
        log_group = 'pipelines_userdata'
        log_filter = 'Command error'
    found_log_group = None
    for url in log_urls:
        if log_group and str(log_group) in url:
            found_log_group = url
    if not found_log_group:
        abort(
            "Log Group '%s' is not available."
            " Choose from:\n  %s" % (log_group, "\n  ".join(log_urls.keys()))
        )
    click.echo("Viewing log group '%s'" % found_log_group)
    url = log_urls[found_log_group]
    if log_filter:
        url += '?filter=%s' % log_filter
    resp = session.get(url)
    if resp.status_code != 200:
        abort(resp.json()['error']['description'])
    logs = resp.text
    if not logs:
        print_warn("No logs found")
        return

    click.echo(logs)
    click.echo("\n%s loglines found" % (logs.count('\n')))


@job.command(help="View events triggered by the job.")
@click.option('--raw', 'is_raw', is_flag=True, help='Dump raw json response from endpoint')
@click.option('--event', '-e', 'event_name', help='Filter by event name')
@click.option('--step', '-s', 'step_name', help='Filter by step name')
@click.pass_context
def events(ctx, is_raw, event_name=None, step_name=None):
    job_resource = ctx.obj.job
    session = ctx.obj.session
    event_url = session.links(job_resource)['events']
    url = event_url
    if event_name or step_name:
        url = event_url + '?'
        if event_name:
            url = url + 'event_name=%s&' % event_name
        if step_name:
            url = url + 'step_name=%s' % step_name

    resp = session.get(url)
    events = resp.json()
    # for backwards compatibility
    if event_name:
        events = [e for e in events if e.get('event_name') == event_name]
    if step_name:
        events = [e for e in events if e['event'].get('step_name') == step_name]
    if is_raw:
        click.echo(dumps(events))
        return
    lst = []
    for event in events:
        dt = dateutil.parser.parse(event['date_time'])
        step_name = event['event'].get('step_name')
        details = event['event'].get('text')
        lst.append([dt.strftime("%H:%M:%S"), event['event_name'], step_name, details])
    click.echo(tabulate(lst, headers=['date', 'event', 'step', 'details']))


@job.command(help="Download all output files into a local folder.")
@click.argument('folder', required=True)
@click.pass_context
def download(ctx, folder):
    job_resource = ctx.obj.job
    session = ctx.obj.session
    files = session.get(session.links(job_resource)['files']).json()
    output_files = files.get('output') or []
    if not output_files:
        click.secho("No output files found", fg="yellow")
    for output in output_files:
        url = output['url']
        file_size = output['file_size'] / 1024 / 1024
        if not url:
            print_warn(
                "Cannot download file '%s' from '%s'. No signed url returned from server"
                % (output['file_name'], output['remote_path'])
            )
        elif file_size < 0:
            print_warn(
                "Cannot download file '%s' from '%s'. File not found"
                % (output['file_name'], output['remote_path'])
            )
        else:
            if file_size <= 0:
                file_size_txt = "<1MB"
            else:
                file_size_txt = "%.0f MB" % file_size
            click.echo(
                "Downloading file '%s' from '%s' (%s)..."
                % (output['file_name'], output['remote_path'], file_size_txt)
            )
            download_output_file(folder, url)


@job.command(help="View job steps and status.")
@click.option('--raw', 'is_raw', is_flag=True, help='Dump raw json response from endpoint')
@click.pass_context
def steps(ctx, is_raw):
    job_resource = ctx.obj.job
    if is_raw:
        click.echo(dumps(job_resource['steps']))
        return

    click.echo(get_steps_table(job_resource))
    if not has_failed_steps(job_resource):
        return

    for step in job_resource['steps']:
        if step['status'] == 'FAILED':
            click.secho(
                "\nStep '%s' (%s) failed with exit code %s. Final loglines from stderr:"
                % (step['step_name'], step['step_identifier'], step['details']['exit_code']),
                fg='red',
            )
            for l in step['details']['error'].split('\n'):
                if l.strip():
                    click.secho(l.rstrip(), fg='red')
    click.echo("\nYou can view additional logs with the 'logs' subcommand.")


@job.command(help="Monitor a running job.")
@click.pass_context
def watch(ctx):
    # count = 100
    # items = range(count)
    # def filter(items):
    #     for item in items:
    #         if random.random() > 0.3:
    #             yield item
    # with click.progressbar(filter(items),
    #                       show_eta=False,
    #                       label='Initializing Worker Instance',
    #                       fill_char=click.style('#', fg='magenta')) as bar:
    #     while 1:
    #         time.sleep(0.02)
    #         bar.update(1)
    job_resource = ctx.obj.job
    session = ctx.obj.session
    job_url = session.links(job_resource)['self']

    try:

        status = job_resource['status']
        if status not in INIT_STATUSES + ('RUNNING',):
            print_warn("Job is not in running state. Nothing to watch.")
            ctx.invoke(view)
            return

        last_status = None
        while status in INIT_STATUSES:
            status = job_resource['status']
            if status != last_status:
                sys.stdout.write("\nJob status: %s." % status)
                sys.stdout.flush()
                last_status = status

            while job_resource['status'] == last_status and last_status in INIT_STATUSES:
                sys.stdout.write(".")
                sys.stdout.flush()
                time.sleep(5.0)
                job_resource = session.get(job_url).json()

        if status == 'RUNNING':
            import curses

            stdscr = curses.initscr()
            curses.noecho()
            try:
                stdscr.clear()
                stdscr.addstr("\nJob is running.")
                while job_resource['status'] == 'RUNNING':
                    stdscr.addstr(".")
                    time.sleep(1.0)
                    job_resource = session.get(job_url).json()
                    stdscr.clear()
                    steps_txt = get_steps_table(job_resource, color=False)
                    try:
                        stdscr.addstr(steps_txt)
                    except Exception:
                        pass
                    stdscr.refresh()
                time.sleep(1.0)
            finally:
                curses.endwin()
            click.echo("\nJob has completed with status: %s" % job_resource['status'])
            click.echo(
                "View with: %s"
                % (
                    click.style(
                        "nextcode pipelines job %s view" % job_resource['job_id'], bold=True
                    )
                )
            )
        else:
            print_error("\nUnexpected job status: %s" % job_resource['status'])
    except KeyboardInterrupt:
        click.echo(
            "\nStopped watching job. To continue run: "
            "nextcode pipelines job %s watch" % job_resource['job_id']
        )
        return


@job.command(help="Duplicate a job and rerun")
@click.pass_context
def duplicate(ctx):
    job_resource = ctx.obj.job
    session = ctx.obj.session
    job_url = session.links(job_resource)['self']
    resp = session.patch(job_url, json={'action': 'duplicate'})
    resp.raise_for_status()
    data = resp.json()
    new_job_id = data['details']['new_job_id']
    click.echo(data['message'])
    click.echo("Please view the new job with nextcode pipelines job %s view" % new_job_id)


from nextcodecli.commands.pipelines.instance import instance

job.add_command(instance)



#!/usr/bin/env python

import os
import requests
from urllib.parse import urlsplit, urljoin

import click
from click.utils import echo

from nextcode.config import (
    root_config_folder,
    get_profiles,
    get_default_profile,
    set_default_profile,
    create_profile,
    delete_profile,
)
from nextcode.exceptions import InvalidProfile
from nextcodecli.utils import abort


@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx):
    """
    Configure server profile to use.

    Note that you can override the currently selected profile with: export NEXTCODE_PROFILE=[profile name]
    """
    if ctx.invoked_subcommand is None:
        ctx.invoke(view, profile=get_default_profile())
        echo(ctx.get_help())


@cli.command()
@click.pass_context
def list(ctx):
    """lists all available profiles."""
    click.echo('Available profiles:')
    for profile_name in sorted(get_profiles()):
        if get_default_profile() == profile_name:
            click.secho('==> %s <==' % profile_name, fg='green')
        else:
            click.echo('    %s' % profile_name)
    click.echo('')


@cli.command()
@click.argument('profile')
@click.pass_context
def view(ctx, profile):
    """displays content for a given profile name."""
    if not get_profiles():
        abort("No profiles are currently set up. Please run 'nextcode profile add [name]' to start")

    profile = (profile or "").lower()
    click.secho('Current profile: %s' % get_default_profile(), fg='green')
    profile_config = get_profiles().get(profile)
    if not profile_config:
        click.secho('Profile %s not found' % profile, fg='red')
        ctx.invoke(list)
        return

    click.echo('Viewing profile configuration for %s' % profile)
    for k, v in sorted(profile_config.items()):
        click.echo("   {0:20}{1}".format(k, v))
    click.echo('')


@cli.command()
@click.argument('profile')
@click.pass_context
def set(ctx, profile):
    """sets the profile given by name as 'current'."""
    profile = profile.lower()
    try:
        set_default_profile(profile)
    except Exception as e:
        click.secho(str(e), fg='red')
        click.echo("View available profiles with 'nextcode profile list'")
        return
    else:
        click.secho('Your profile has been changed to %s' % profile, bold=True)

    ctx.invoke(view, profile=profile)


@cli.command()
@click.pass_context
def edit(ctx):
    """open up the config.yaml file"""
    full_filename = os.path.join(root_config_folder, "config.yaml")
    click.echo("Editing config file located at %s" % (full_filename))

    click.echo("Modify your config and close the file")
    click.edit(filename=full_filename, require_save=True)


@cli.command()
@click.argument('profile')
@click.option(
    '-d', '--domain', prompt="Domain name of the service (f.ex. platform.wuxinextcodedev.com)"
)
@click.option('-k', '--api-key', help="API Key (if not set a browser will open)")
@click.option('-r', '--replace', is_flag=True, default=False, help="Replace an existing profile")
@click.option(
    '-s',
    '--skip-auth',
    is_flag=True,
    default=False,
    help="Skip authenetication (for local development)",
)
@click.pass_context
def add(ctx, profile, domain, api_key, replace, skip_auth):
    """Creates a new profile of the specified name and allows you to configure it"""
    profile = profile.lower()
    if profile in get_profiles() and not replace:
        click.secho(
            "Profile '{}' already exists. Use --replace if you want to override it.".format(
                profile
            ),
            fg='red',
        )
        return

    if not domain.startswith("http"):
        domain = "https://{}".format(domain)
    # check for common mistakes in domain name
    if not domain or ':' in domain or '/' in domain:
        parts = urlsplit(domain)
        domain = "{}://{}".format(parts.scheme, parts.netloc)

    if not skip_auth:
        if not api_key:
            reachable = True
            try:
                url = '%s/api-key-service' % domain
                resp = requests.get(url, timeout=2)
                resp.raise_for_status()
            except Exception:
                click.secho("%s is not reachable." % url, fg='red')
                reachable = False
                if not click.confirm("Are you sure you want to add this domain?"):
                    return

            if reachable:
                click.echo(
                    "Launching login webpage {} ==> Please retrieve API Key from browser".format(
                        url
                    )
                )
                click.launch(url)
                click.pause()
            else:
                click.secho("Please aquire an api_key from " "somewhere manually.", fg='yellow')
            # Note: readline must be imported for click.prompt() to accept long strings. Don't ask me why.
            import readline

            api_key = click.prompt("API Key", type=str)

    try:
        create_profile(profile, api_key=api_key, root_url=domain, skip_auth=skip_auth)
    except InvalidProfile as ex:
        abort(ex)
    set_default_profile(profile)
    click.echo(
        "Profile has been added and is now active. "
        "If you want to change it please run: nextcode profile edit"
    )


@cli.command()
@click.argument('profile')
@click.pass_context
def delete(ctx, profile):
    """accepts a profile name and deletes it."""
    profile = profile.lower()
    delete_profile(profile)
    click.secho("Profile '%s' has been deleted" % profile, bold=True)


def custom_error(main_command):
    def show(self, file=None):
        click.secho('Current profile: %s' % get_default_profile(), fg='green')
        click.echo(main_command.get_help(self.ctx))

    click.exceptions.UsageError.show = show


custom_error(cli)



#!/usr/bin/env python

import click
import logging
import nextcodecli

log = logging.getLogger(__name__)


@click.command()
@click.pass_context
def cli(ctx):
    """
    Show the Nextcode CLI version.
    """
    click.echo(nextcodecli.get_version_string())



#!/usr/bin/env python
import os, sys
from setuptools import setup, find_packages
# https://stackoverflow.com/questions/49837301/pip-10-no-module-named-pip-req
try:  # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError:  # for pip <= 9.0.3
    from pip.req import parse_requirements

if sys.version_info < (3, 6):
    raise NotImplementedError(
        """Nextcode SDK does not support Python versions older than 3.6"""
    )

root_dir = 'nextcodecli'


def get_version_string():
    with open(os.path.join(root_dir, 'VERSION')) as version_file:
        return version_file.readlines()[0].strip()


version = get_version_string()
if 'SETUP_BRANCH' in os.environ:
    version += "-%s" % os.environ['SETUP_BRANCH']

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='nextcode-cli',
    python_requires=">=3.6",
    version=version,
    description="WuXi Nextcode commandline utilities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='WUXI NextCODE',
    author_email='support@wuxinextcode.com',
    url='https://www.wuxinextcode.com',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    package_data={'nextcodecli': ['VERSION', 'PUBLIC_KEY']},
    install_requires=[
        str(i.req)
        for i in parse_requirements('requirements.txt', session=False)
        if i.req
    ],
    entry_points={
        'console_scripts': [
            'nextcode = nextcodecli.__main__:cli',
        ]
    },
)



#!/usr/bin/env python

import botocore.session
import click
import configparser
import logging
import os
import sys
import json
from jsonpath_rw import parse
import yaml

log_formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

console_handler = logging.StreamHandler()
console_handler.setFormatter(log_formatter)
console_handler.setLevel(logging.WARN)

root_logger = logging.getLogger()
root_logger.setLevel(logging.DEBUG)
root_logger.addHandler(console_handler)


def print_error(*txt):
    color_print(*txt, color='red')


def print_warn(*txt):
    color_print(*txt, color='yellow')


def print_success(*txt):
    color_print(*txt, color='green')


def color_print(*args, **kwargs):
    color = kwargs.get('color', 'reset')
    for x in args:
        click.echo(click.style(x, fg=color))


def print_tab(key, val):
    from builtins import str  # py2 backwards compatibility

    val = click.style(str(val), bold=True)
    click.echo("{0:20}{1}".format(key, val))


def print_table(obj):
    try:
        for k, v in obj.items():
            print_tab(k, v)
    except Exception:
        click.echo(obj)


def dumps(obj):
    try:
        return json.dumps(obj, indent=4, sort_keys=True)
    except Exception:
        return repr(obj)


def output_pretty_json(dct, keys):
    ret = ""
    for k in keys:
        jsonpath_expr = parse(k)
        ret = [match.value for match in jsonpath_expr.find(dct)]
        lst = ['']
        if ret:
            if isinstance(ret[0], list) or isinstance(ret[0], dict):
                lst = yaml.dump(ret[0]).split('\n')
            else:
                lst = [str(ret[0])]
        print_tab(k.split('.')[-1], lst[0])
        if len(lst) > 1:
            for v in lst[1:]:
                print_tab("", v)


def get_logger(name='nextcodecli', level=None):
    # TODO - refactor this funtion away - it's useless
    return logging.getLogger(name)


def find_aws_credentials(profile):
    """
    Returns the aws credentials for the specified profile.
    If no profile is passed in, returns the credentials for the currently selected profile

    Args:
        profile name

    Returns:
        Dict containing at least aws_access_key_id, aws_secret_access_key

    Raises:
        RuntimeError is no default profile or the named profile was not found

    """
    if not profile:
        access_key = None
        secret_key = None
        token = ''
        credentials = botocore.session.get_session().get_credentials()
        if credentials:
            access_key = credentials.access_key
            secret_key = credentials.secret_key
            token = getattr(credentials, 'token') or ''
        if not access_key or not secret_key:
            raise RuntimeError("No Default AWS profile set")

        return {
            "aws_access_key_id": access_key,
            "aws_secret_access_key": secret_key,
            "aws_session_token": token,
        }

    folder = os.path.join(os.path.expanduser("~"), ".aws")
    filename = os.path.join(folder, 'credentials')
    cfg = configparser.ConfigParser()
    with open(filename) as fp:
        cfg.readfp(fp)
        ret = {}
        if profile not in cfg:
            raise RuntimeError("No AWS profile '%s' found in %s" % (profile, filename))
        for key in cfg[profile]:
            ret[key] = cfg[profile][key]
    return ret


def abort(reason, code=1):
    """
    exit with non-zero exit code and write reason for error to stderr.
    If we are outside a click context and exception will be raised instead
    """
    click.echo(click.style(str(reason), fg='red'), file=sys.stderr)
    ctx_test = click.get_current_context(silent=True)
    if ctx_test:
        sys.exit(code)
    else:
        raise Exception("Abnormal Termination")


def check_profile(ctx):
    if not ctx.obj.client:
        abort(
            "You have no active profile or selected profile is invalid. Run 'nextcode profile add [name]'"
        )



#!/usr/bin/env python

from click import echo, secho, pass_context, command, argument, option
import dateutil
from tabulate import tabulate
import json

from nextcodecli.utils import dumps, print_tab


@command(help="List pipelines")
@argument('pipeline_name', nargs=-1)
@option('--raw', 'is_raw', is_flag=True, help='Dump raw json for further processing')
@option('-d', '--details', 'is_details', is_flag=True, help='View details about each pipeline')
@option('-m', '--manifest', 'is_manifest', is_flag=True, help='Output a manifest for this pipeline')
@pass_context
def pipelines(ctx, pipeline_name, is_raw, is_details, is_manifest):
    pipelines = ctx.obj.service.get_pipelines()

    _pipeline_names = set()
    if pipeline_name:
        for p in pipelines:
            for pn in pipeline_name:
                if p['name'].startswith(pn):
                    _pipeline_names.add(p['name'])
        pipelines = [p for p in pipelines if p['name'] in _pipeline_names]
        is_details = True

    if is_manifest:
        for pipeline in pipelines:
            manifest = {
                "pipeline_name": pipeline['name'],
                "project_name": "<project_name>",
                "script": None,
                "revision": None,
                "profile": None,
                "params": {},
            }
            for i, p in enumerate(pipeline.get('parameters', [])):
                val = None
                if p.get('required'):
                    val = "<%s>" % p['name']
                manifest['params'][p['name']] = val
            echo(json.dumps(manifest, indent=4))
    elif is_raw:
        echo(dumps(pipelines))
        return

    elif is_details:
        for p in pipelines:
            print_tab("Pipeline Name", p['name'])
            print_tab("Description", p['description'])
            print_tab("Script", p['script'])
            print_tab("Revision", p['revision'] or 'latest')
            for i, p in enumerate(p.get('parameters', [])):
                txt = "%s (%s)" % (p['name'], p['type'])
                if p.get('required'):
                    txt += " - required"
                if p.get('description'):
                    txt += "\t # %s" % (p.get('description'))
                k = "parameters" if i == 0 else ""
                print_tab(k, txt)
            print(" ")
    else:
        fields = ['name', 'script', 'revision']
        table = []
        for pipeline in pipelines:
            lst = []
            for f in fields:
                v = pipeline.get(f) or 'N/A'
                lst.append(v)

            table.append(lst)
        tbl = tabulate(table, headers=fields)
        echo(tbl)
